@@ -20,6 +20,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <stdlib.h>
 #include <string.h>
 #include <popt.h>
 #include <gdk/gdkwindow.h>
@@ -51,7 +52,11 @@ enum {
 	ZOOM_REGION_XALIGN_PROP,
 	ZOOM_REGION_YALIGN_PROP,
 	ZOOM_REGION_VIEWPORT_PROP,
-	ZOOM_REGION_TESTPATTERN_PROP
+	ZOOM_REGION_TESTPATTERN_PROP,
+	ZOOM_REGION_TIMING_TEST_PROP,
+	ZOOM_REGION_TIMING_OUTPUT_PROP,
+	ZOOM_REGION_TIMING_PAN_RATE_PROP,
+	ZOOM_REGION_EXIT_MAGNIFIER
 } PropIdx;
 
 typedef enum {
@@ -60,6 +65,102 @@ typedef enum {
 	ZOOM_REGION_ERROR_TOO_BIG
 } ZoomRegionPixmapCreationError;
 
+/*
+ * When running the magnifier with the panning test (--timing-pan-rate), the following
+ * output is displayed:
+ *
+ * ### Updates summary:
+ * 
+ *   Update Duration          = (avg. 0.010232) (max. 0.002685) (tot. 1.084589) seconds
+ *     Update Pixels          = (avg. 3046) pixels/frame
+ *     Update Rate            = (avg. 97.732909) (max. 372.439485) updates/second
+ *     Net Update Rate        = (avg. 0.297724) (max. 0.457741) Mpex/second
+ * 
+ * ### Updates finished, starting panning test
+ * 
+ *   Pan Rate                 = 200
+ *
+ * Final Report:
+ *
+ *   Frames Processed         = 1225
+ *   Width/Height/Depth       = 640/1024/8
+ *   Zoom Factor (x/y)        = 3.000000/3.000000
+ *   Pan Latency              = (avg. 0.008229) (max. 0.562326) seconds
+ *   Total Frame Duration     = (avg. 0.008363) (max. 0.562485) (tot. 10.235981) seconds
+ *   Frame Rate               = (avg. 119.578189) (max. 4854.369141) frames/second
+ *   Scroll Delta (x)         = (avg. 0.000000) (tot. 0) lines
+ *   Scroll Delta (y)         = (avg. 5.019608) (tot. 6144) lines
+ *   Scroll Rate (x)          = (avg. 0.000000) lines/second
+ *   Scroll Rate (y)          = (avg. 600.235657) lines/second
+ *   Net Render Rate          = (avg. 8.685322) (max. 3181.359360) Mpex/second
+ * 
+ * Pan Rate             - Displays the pan rate specified on the command line.
+ * Frames Processed     - Number of frames displayed
+ * Width/Height/Depth   - The width/height/depth of magnifier display.
+ * Zoom Factor (x/y)    - The zoom factor used.
+ * Pan Latency          - Time spent from the beginning of the setROI request until GTK+
+ *                        has finished processing the update (in GDK's 
+ *                        gdk_window_update_idle idle handler.
+ * Total Frame Duration - Average and total time processing frames.
+ * Frame Rate           - Number of frames per second.
+ * Scroll Delta         - Average and total number of lines scrolled.
+ * Scroll Rate          - Number of lines scrolled per second.
+ * Net Render Rate      - Number of Megapixels processed per second.
+ * 
+ * Note that updates do not happen for some frames, such as when panning.  When updates
+ * do happen, the following lines are included:
+ *
+ * Update Duration   - Average and total time spent updating new sections of the screen.
+ *   Update Pixels   - Number of pixels updated
+ *   Update Rate     - Number of updates possible per second.
+ *   Net Update Rate - Number of Megapixels updated per second.
+ *
+ * When running with the --timing-output argument, similar information is displayed for
+ * each frame plus timing information for the frame.
+ */ 
+
+typedef struct timing {
+	struct timeval process_start;
+	struct timeval process_end;
+	struct timeval frame_start;
+	struct timeval frame_end;
+	struct timeval scale_start;
+	struct timeval scale_end;
+	struct timeval idle_start;
+	struct timeval idle_end;
+	float scale_val;
+	float idle_val;
+	float frame_val;
+	int   dx;
+	int   dy;
+	float scale_total;
+	float idle_total;
+	float frame_total;
+	int   dx_total;
+	int   dy_total;
+	long  update_pixels_total;
+	long  num_scale_samples;
+	long  num_idle_samples;
+	long  num_frame_samples;
+	long  num_line_samples;
+	float last_frame_val;
+	int   last_dy;
+} TimingStruct;
+
+static float timing_scale_max  = 0;
+static float timing_idle_max   = 0;
+static float timing_frame_max  = 0;
+static float cps_max           = 0;
+static float nrr_max           = 0;
+static float update_nrr_max    = 0;
+static gboolean	reset_timing   = FALSE;
+static gboolean timing_test    = FALSE;
+
+static guint pending_idle_handler = 0;
+static TimingStruct mag_timing;
+static gboolean processing_updates = FALSE;
+static gboolean timing_start = FALSE;
+
 #ifdef TEST_XTST_CURSOR
 static Cursor *x_cursors;
 static Window cursor_window = None;
@@ -79,6 +180,25 @@ static void zoom_region_paint (ZoomRegion *zoom_region, GdkRectangle *rect);
 static void zoom_region_paint_pixmap (ZoomRegion *zoom_region, GdkRectangle *rect);
 static int  zoom_region_update_pointer_timeout (gpointer data);
 
+void
+reset_timing_stats()
+{
+	mag_timing.num_scale_samples   = 0;
+	mag_timing.num_idle_samples    = 0;
+	mag_timing.num_frame_samples   = 0;
+	mag_timing.num_line_samples    = 0;
+	mag_timing.scale_total         = 0;
+	mag_timing.idle_total          = 0;
+	mag_timing.frame_total         = 0;
+	mag_timing.update_pixels_total = 0;
+	mag_timing.update_pixels_total = 0;
+	mag_timing.dx_total            = 0;
+	mag_timing.dy_total            = 0;
+	mag_timing.last_frame_val      = 0;
+	mag_timing.last_dy             = 0;
+	gettimeofday(&mag_timing.process_start, NULL);
+}
+
 /** DEBUG STUFF **/
 
 #undef DEBUG
@@ -465,7 +585,7 @@ zoom_region_clip_to_exposed_target (ZoomRegion *zoom_region,
 				  zoom_region->priv->source_area.x2);
 	onscreen_target.y2 = MIN (zoom_region->priv->exposed_bounds.y2 / zoom_region->yscale,
 				  zoom_region->priv->source_area.y2);
-	
+
 	return _rectangle_clip_to_bounds (area, &onscreen_target);
 }
 
@@ -1014,7 +1134,10 @@ zoom_region_scroll_fast (ZoomRegion *zoom_region, int dx, int dy,
 	GdkWindow *window;
 	if (zoom_region->priv->w && zoom_region->priv->w->window)
 		window = zoom_region->priv->w->window;
-	else return;
+	else {
+		processing_updates = FALSE;
+		return;
+	}
 	if (zoom_region->smooth_scroll_policy > GNOME_Magnifier_ZoomRegion_SCROLL_FAST) 
 		gdk_display_sync (gdk_drawable_get_display (window));
 	gdk_window_scroll (window, dx, dy);
@@ -1053,6 +1176,27 @@ zoom_region_scroll (ZoomRegion *zoom_region, int dx, int dy)
 {
 	GdkRectangle scroll_rect, expose_rect_h, expose_rect_v;
 	gboolean can_scroll;
+
+	if (timing_test) {
+		mag_timing.num_line_samples++;
+		mag_timing.dx = abs(dx);
+		mag_timing.dy = abs(dy);
+		mag_timing.dx_total += mag_timing.dx;
+		mag_timing.dy_total += mag_timing.dy;
+		if (zoom_region->timing_output) {
+			fprintf(stderr, "  Panning Increment (x)    = %d (avg. %f) lines/frame\n",
+				mag_timing.dx, (float)mag_timing.dx_total / (float)mag_timing.num_line_samples);
+			fprintf(stderr, "  Panning Increment (y)    = %d (avg. %f) lines/frame\n",
+				mag_timing.dy, (float)mag_timing.dy_total / (float)mag_timing.num_line_samples);
+		}
+	}
+
+    /*
+     * Currently processing a screen update.  This flag used to disallow
+     * other updates to occur until this one finishes
+     */
+    processing_updates = TRUE;
+
 	can_scroll = zoom_region_calculate_scroll_rects (zoom_region, dx, dy,
 							 &scroll_rect,
 							 &expose_rect_h,
@@ -1137,9 +1281,9 @@ zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 	gint mouse_x_return, mouse_y_return;
 	guint mask_return;
 
-	if (!zoom_region->priv) 
+	if (!zoom_region->priv || !zoom_region->priv->parent)
 	  return FALSE; 
-	if (!zoom_region->priv->parent) return FALSE;
+
 	magnifier = zoom_region->priv->parent;
 	/* TODO: there's really no reason we should be using magnifier->priv->root here */
 	if (magnifier && magnifier->priv && magnifier_get_root (magnifier))
@@ -1267,6 +1411,10 @@ zoom_region_moveto (ZoomRegion *zoom_region,
 	long dx = x * zoom_region->xscale - zoom_region->priv->exposed_bounds.x1;
 	long dy = y * zoom_region->yscale - zoom_region->priv->exposed_bounds.y1;
 /*	fprintf (stderr, "moveto %ld %ld\n", x, y); */
+
+	mag_timing.dx = 0;
+	mag_timing.dy = 0;
+
 	if ((dx != 0) || (dy != 0)) {
 		zoom_region_update_pointer (zoom_region, FALSE);
 		zoom_region->priv->exposed_bounds.x1 = x * zoom_region->xscale;
@@ -1435,6 +1583,7 @@ zoom_region_update (ZoomRegion *zoom_region,
 	if (subimage && zoom_region->priv->w && zoom_region->priv->w->window)
 	{
 		GdkRectangle paint_rect;
+		gettimeofday(&mag_timing.scale_start, NULL);
 		paint_rect = zoom_region_view_rect_from_source_rect (zoom_region, source_rect);
 		paint_rect = zoom_region_clip_to_window (zoom_region, paint_rect);
 		DEBUG_RECT ("paint rect", paint_rect);
@@ -1476,6 +1625,37 @@ zoom_region_update (ZoomRegion *zoom_region,
 		gdk_window_begin_paint_rect (zoom_region->priv->w->window, &paint_rect);
 		zoom_region_paint (zoom_region, &paint_rect);
 		gdk_window_end_paint (zoom_region->priv->w->window);
+   		gettimeofday(&mag_timing.scale_end, NULL);
+		if (timing_test) {
+			mag_timing.num_scale_samples++;
+			mag_timing.scale_val =
+				(mag_timing.scale_end.tv_sec - mag_timing.scale_start.tv_sec) +
+				((float)(mag_timing.scale_end.tv_usec - mag_timing.scale_start.tv_usec) /
+				1000000.0);
+			mag_timing.scale_total += mag_timing.scale_val;
+
+			if (mag_timing.scale_val != 0 && (timing_scale_max == 0 ||
+			   (1.0/(float)mag_timing.scale_val) > (1.0/(float)timing_scale_max)))
+				timing_scale_max = mag_timing.scale_val;
+			if ((source_rect.height * source_rect.width / mag_timing.scale_val) > update_nrr_max)
+				update_nrr_max = source_rect.height * source_rect.width / mag_timing.scale_val;
+
+			mag_timing.update_pixels_total += source_rect.height * source_rect.width;
+
+			if (zoom_region->timing_output) {
+				fprintf(stderr, "  Update Duration          = %f (avg. %f) (max. %f) (tot. %f) seconds\n",
+					mag_timing.scale_val, (mag_timing.scale_total / 
+					mag_timing.num_scale_samples), timing_scale_max, mag_timing.scale_total);
+				fprintf(stderr, "    Update Pixels          = %ld (avg. %ld) pixels/frame\n",
+					source_rect.height * source_rect.width,
+					mag_timing.update_pixels_total / mag_timing.num_scale_samples);
+				fprintf(stderr, "    Update Rate            = (avg. %f) (max. %f) updates/second\n",
+					1.0/(mag_timing.scale_total / mag_timing.num_scale_samples), 1.0/(float)timing_scale_max);
+				fprintf(stderr, "    Net Update Rate        = (avg. %f) (max. %f) Mpex/second\n",
+					((float)mag_timing.update_pixels_total / (float)mag_timing.scale_total) / 1000000.0,
+					update_nrr_max / 1000000.0);
+			}
+		}
 	} else {
 		if (subimage)
 			fprintf (stderr, "update on uninitialized zoom region!\n");
@@ -1539,7 +1719,128 @@ zoom_region_process_updates (gpointer data)
 		return TRUE;
 	}
 	else 
-	        return FALSE;
+		return FALSE;
+}
+
+static void
+zoom_region_time_frame(ZoomRegion *zoom_region, Magnifier *magnifier)
+{
+	float frame_avg;
+	float x_scroll_incr, y_scroll_incr;
+	int width = magnifier->target_bounds.x2 - magnifier->target_bounds.x1;
+	int height = magnifier->target_bounds.y2 - magnifier->target_bounds.y1;
+
+	mag_timing.num_frame_samples++;
+	gettimeofday(&mag_timing.frame_end, NULL);
+	mag_timing.frame_val = (mag_timing.frame_end.tv_sec - mag_timing.frame_start.tv_sec) +
+		((float)(mag_timing.frame_end.tv_usec - mag_timing.frame_start.tv_usec) /
+		1000000.0);
+
+	mag_timing.frame_total += mag_timing.frame_val;
+	if (mag_timing.frame_val > timing_frame_max)
+		timing_frame_max = mag_timing.frame_val;
+	if (mag_timing.frame_val != 0 && 1.0/mag_timing.frame_val > cps_max)
+		cps_max = 1.0/mag_timing.frame_val;
+
+	frame_avg = mag_timing.frame_total / mag_timing.num_frame_samples;
+
+	x_scroll_incr = (float)mag_timing.dx_total / (float)mag_timing.num_line_samples;
+	y_scroll_incr = (float)mag_timing.dy_total / (float)mag_timing.num_line_samples;
+
+	if ((height * width / mag_timing.frame_val) > nrr_max)
+		nrr_max = height * width / mag_timing.frame_val;
+
+	if (zoom_region->timing_output) {
+		fprintf(stderr, "  Total Frame Duration     = %f (avg. %f) (max. %f) (tot. %f) seconds\n",
+			mag_timing.frame_val, frame_avg, timing_frame_max, mag_timing.frame_total);
+		fprintf(stderr, "  Frame Rate               = (avg. %f) (max. %f) frames/second\n",
+			1.0 /frame_avg, cps_max);
+		fprintf(stderr, "  Scroll Delta (x)         = (avg. %f) (tot. %d) lines\n",
+			x_scroll_incr, mag_timing.dx_total);
+		fprintf(stderr, "  Scroll Delta (y)         = (avg. %f) (tot. %d) lines\n",
+			y_scroll_incr, mag_timing.dy_total);
+		fprintf(stderr, "  Scroll Rate (x)          = (avg. %f) lines/second\n",
+			x_scroll_incr / frame_avg);
+		fprintf(stderr, "  Scroll Rate (y)          = (avg. %f) lines/second\n",
+			y_scroll_incr / frame_avg);
+
+		fprintf(stderr, "  Net Render Rate          = (avg. %f) (max. %f) Mpex/second\n",
+			(height * width *
+			((float)mag_timing.num_frame_samples / (float)mag_timing.frame_total)) / 1000000.0,
+			nrr_max / 1000000.0);
+	}
+
+	mag_timing.last_frame_val = mag_timing.frame_val;
+	mag_timing.last_dy        = mag_timing.dy;
+
+	if (reset_timing) {
+		fprintf(stderr, "\n### Updates summary:\n\n");
+		fprintf(stderr, "  Update Duration          = (avg. %f) (max. %f) (tot. %f) seconds\n",
+			(mag_timing.scale_total / mag_timing.num_scale_samples), timing_scale_max, mag_timing.scale_total);
+		fprintf(stderr, "    Update Pixels          = (avg. %ld) pixels/frame\n",
+			mag_timing.update_pixels_total / mag_timing.num_scale_samples);
+		fprintf(stderr, "    Update Rate            = (avg. %f) (max. %f) updates/second\n",
+			1.0/((float)mag_timing.scale_total / (float)mag_timing.num_scale_samples),
+			1.0/(float)timing_scale_max);
+		fprintf(stderr, "    Net Update Rate        = (avg. %f) (max. %f) Mpex/second\n",
+			((float)mag_timing.update_pixels_total / (float)mag_timing.scale_total) / 1000000.0,
+			update_nrr_max / 1000000.0);
+	    	fprintf(stderr, "\n### Updates finished, starting panning test\n");
+		reset_timing_stats();
+		reset_timing = FALSE;
+	}
+}
+
+void
+exit_report(ZoomRegion *zoom_region)
+{
+	float frame_avg;
+	float x_scroll_incr, y_scroll_incr;
+	int width, height, x, y;
+
+	if (timing_test) {
+		width = (zoom_region->priv->exposed_viewport.x2 -
+			zoom_region->priv->exposed_viewport.x1) / zoom_region->xscale;
+		height = (zoom_region->priv->exposed_viewport.y2 -
+			zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
+
+		frame_avg = mag_timing.frame_total / mag_timing.num_frame_samples;
+
+		x_scroll_incr = (float)mag_timing.dx_total / (float)mag_timing.num_line_samples;
+		y_scroll_incr = (float)mag_timing.dy_total / (float)mag_timing.num_line_samples;
+
+		gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
+			&x, &y);
+
+		fprintf(stderr, "\nFinal Report:\n\n");
+
+		fprintf(stderr, "  Frames Processed         = %d\n", 
+			mag_timing.num_frame_samples + 1);
+		fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
+			gdk_drawable_get_depth (zoom_region->priv->w->window));
+		fprintf(stderr, "  Zoom Factor (x/y)        = %f/%f\n", zoom_region->xscale,
+			zoom_region->yscale);
+		fprintf(stderr, "  Pan Latency              = (avg. %f) (max. %f) seconds\n",
+			(mag_timing.idle_total / mag_timing.num_idle_samples), timing_idle_max);
+		fprintf(stderr, "  Total Frame Duration     = (avg. %f) (max. %f) (tot. %f) seconds\n",
+			frame_avg, timing_frame_max, mag_timing.frame_total);
+		fprintf(stderr, "  Frame Rate               = (avg. %f) (max. %f) frames/second\n",
+			1.0 / (mag_timing.frame_total / mag_timing.num_frame_samples), cps_max);
+		fprintf(stderr, "  Scroll Delta (x)         = (avg. %f) (tot. %d) lines\n",
+			x_scroll_incr, mag_timing.dx_total);
+		fprintf(stderr, "  Scroll Delta (y)         = (avg. %f) (tot. %d) lines\n",
+			y_scroll_incr, mag_timing.dy_total);
+		fprintf(stderr, "  Scroll Rate (x)          = (avg. %f) lines/second\n",
+			x_scroll_incr / frame_avg);
+		fprintf(stderr, "  Scroll Rate (y)          = (avg. %f) lines/second\n",
+			y_scroll_incr / frame_avg);
+
+		fprintf(stderr, "  Net Render Rate          = (avg. %f) (max. %f) Mpex/second\n\n",
+			(height * width *
+			((float)mag_timing.num_frame_samples / (float)mag_timing.frame_total)) / 1000000.0,
+			nrr_max / 1000000.0);
+	}
+	exit(0);
 }
 
 static void
@@ -1549,16 +1850,101 @@ zoom_region_sync (ZoomRegion *zoom_region)
 		zoom_region_process_updates (zoom_region);
 }
 
+static gboolean
+gdk_timing_idle (gpointer data)
+{
+	ZoomRegion *zoom_region = data;
+
+	/* Now update has finished, reset processing_updates */
+	processing_updates = FALSE;
+	gettimeofday(&mag_timing.idle_end, NULL);
+
+	if (timing_test) {
+		mag_timing.num_idle_samples++;
+		mag_timing.idle_val =
+			(mag_timing.idle_end.tv_sec - mag_timing.idle_start.tv_sec) +
+			((float)(mag_timing.idle_end.tv_usec - mag_timing.idle_start.tv_usec) /
+			1000000.0);
+		mag_timing.idle_total += mag_timing.idle_val;
+
+		if (mag_timing.idle_val > timing_idle_max)
+			timing_idle_max = mag_timing.idle_val;
+
+		if (zoom_region->timing_output) {
+			fprintf(stderr, "  Pan Latency              = %f (avg. %f) (max. %f) seconds\n",
+				mag_timing.idle_val, (mag_timing.idle_total /
+				mag_timing.num_idle_samples), timing_idle_max);
+		}
+	}
+
+	return FALSE;
+}
+
 static void
 zoom_region_align (ZoomRegion *zoom_region)
 {
+	Magnifier *magnifier = zoom_region->priv->parent;
 	long x = 0, y = 0;
 	long width, height;
 
+	if (timing_start)
+		zoom_region_time_frame(zoom_region, magnifier);
+
+	if (timing_test) {
+		gettimeofday(&mag_timing.frame_start, NULL);
+
+		if (zoom_region->timing_output) {
+			Magnifier *magnifier = zoom_region->priv->parent;
+			gint x, y;
+
+			gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
+				&x, &y);
+
+			fprintf(stderr, "\nTiming Information - ROI   = (%d, %d) (%d, %d):\n",
+				zoom_region->roi.x1, zoom_region->roi.y1, zoom_region->roi.x2,
+				zoom_region->roi.y2);
+			fprintf(stderr, "  Frame Number             = %d\n", 
+				mag_timing.num_frame_samples + 1);
+			fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
+				gdk_drawable_get_depth (zoom_region->priv->w->window));
+		}
+
+		/*
+		 * The timing_start flag makes sure that we don't start displaying output
+		 * until we have processed an entire frame.
+		 */
+		if (!timing_start)
+			gettimeofday(&mag_timing.process_start, NULL);
+
+		timing_start = TRUE;
+	}
+
+	gettimeofday(&mag_timing.idle_start, NULL);
+
+	/*
+	 * zoom_region_align calls
+	 *   zoom_region_moveto calls
+	 *     zoom_region_scroll calls
+	 *        zoom_region_scroll_fast or zoom_region_scroll_smooth calls
+	 *           gdk_window_scroll or gdk_window_invalidate_rect calls
+	 *              gdk_window_invalidate_region calls
+	 *                 gdk_window_invalidate_maybe_recurse
+	 * 
+	 * The last function in the stack will set up an idle handler of
+	 * priority GDK_PRIORITY_REDRAW (gdk_window_update_idle) to be called
+	 * to handle the work of updateing the screen.
+	 *
+	 * By setting up an idle handler of priority GDK_PRIORITY_REDRAW + 1,
+	 * it will be called immediately after and we can determine when GTK+
+	 * is finished with the update.
+	 */
+	g_idle_add_full (GDK_PRIORITY_REDRAW + 1,
+		gdk_timing_idle, zoom_region, NULL);
+
 	width = (zoom_region->priv->exposed_viewport.x2 -
-		 zoom_region->priv->exposed_viewport.x1) / zoom_region->xscale;
+		zoom_region->priv->exposed_viewport.x1) / zoom_region->xscale;
 	height = (zoom_region->priv->exposed_viewport.y2 -
-		  zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
+		zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
 
 	switch (zoom_region->x_align_policy) {
 	case GNOME_Magnifier_ZoomRegion_ALIGN_MAX:
@@ -1669,6 +2055,18 @@ zoom_region_get_property (BonoboPropertyBag *bag,
 					GNOME_Magnifier_RectBounds,
 					NULL);
 		break;
+	case ZOOM_REGION_TIMING_TEST_PROP:
+		BONOBO_ARG_SET_INT (arg, zoom_region->timing_iterations);
+		break;
+	case ZOOM_REGION_TIMING_OUTPUT_PROP:
+		BONOBO_ARG_SET_BOOLEAN (arg, zoom_region->timing_output);
+		break;
+	case ZOOM_REGION_TIMING_PAN_RATE_PROP:
+		BONOBO_ARG_SET_INT (arg, zoom_region->timing_pan_rate);
+		break;
+	case ZOOM_REGION_EXIT_MAGNIFIER:
+		BONOBO_ARG_SET_BOOLEAN (arg, zoom_region->exit_magnifier);
+		break;
 	default:
 		bonobo_exception_set (ev, ex_Bonobo_PropertyBag_NotFound);
 	};
@@ -1755,11 +2153,131 @@ zoom_region_set_property (BonoboPropertyBag *bag,
 						 NULL);
 		zoom_region_set_viewport (zoom_region, &bounds);
 		break;
+	case ZOOM_REGION_TIMING_TEST_PROP:
+		zoom_region->timing_iterations = BONOBO_ARG_GET_INT (arg);
+		timing_test = TRUE;
+		fprintf (stderr, "Setting timing iterations property to %d\n", zoom_region->timing_iterations);
+		break;
+	case ZOOM_REGION_TIMING_OUTPUT_PROP:
+		zoom_region->timing_output = BONOBO_ARG_GET_BOOLEAN (arg);
+		if (zoom_region->timing_output)
+			fprintf (stderr, "Setting timing output property to true\n");
+		else
+			fprintf (stderr, "Setting timing output property to false\n");
+		break;
+	case ZOOM_REGION_TIMING_PAN_RATE_PROP:
+		zoom_region->timing_pan_rate = BONOBO_ARG_GET_INT (arg);
+		timing_test = TRUE;
+		fprintf (stderr, "Setting timing pan rate property to %d\n", zoom_region->timing_pan_rate);
+		break;
+	case ZOOM_REGION_EXIT_MAGNIFIER:
+		zoom_region->exit_magnifier = BONOBO_ARG_GET_BOOLEAN (arg);
+		if (zoom_region->exit_magnifier)
+			fprintf (stderr, "Setting exit magnifier property to true\n");
+		else
+			fprintf (stderr, "Setting exit magnifier property to false\n");
+		break;
 	default:
 		bonobo_exception_set (ev, ex_Bonobo_PropertyBag_NotFound);
 	};
 }
 
+static int
+zoom_region_process_pending (gpointer data)
+{
+	ZoomRegion *zoom_region = (ZoomRegion *) data;
+	Magnifier *magnifier = zoom_region->priv->parent;
+
+	zoom_region_align (zoom_region);
+	return FALSE;
+}
+
+static int
+zoom_region_pan_test (gpointer data)
+{
+	ZoomRegion *zoom_region = (ZoomRegion *) data;
+	Magnifier *magnifier = zoom_region->priv->parent; 
+	GNOME_Magnifier_ZoomRegionList *zoom_regions;
+	GNOME_Magnifier_RectBounds roi;
+	CORBA_Environment ev;
+	static int counter = 0;
+	static gboolean last_update = !TRUE;
+	static gboolean finished_update = !TRUE;
+	float pixels_at_speed;
+	float total_time;
+	int screen_height, height, width;
+	int pixel_position;
+	int pixel_direction;
+
+	screen_height = gdk_screen_get_height (
+		gdk_display_get_screen (magnifier->source_display,
+		 magnifier->source_screen_num));
+
+	height = (zoom_region->priv->exposed_viewport.y2 -
+		zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
+
+	roi.x1 = zoom_region->roi.x1;
+	roi.x2 = zoom_region->roi.x2;
+
+	gettimeofday(&mag_timing.process_end, NULL);
+	total_time = (mag_timing.process_end.tv_sec - mag_timing.process_start.tv_sec) +
+		((float)(mag_timing.process_end.tv_usec - mag_timing.process_start.tv_usec) /
+		1000000.0);
+
+	if (mag_timing.frame_total != 0.0)
+		pixels_at_speed = total_time * zoom_region->timing_pan_rate;
+	else
+		pixels_at_speed = 0.0;
+
+	pixel_position = (int)(pixels_at_speed) % (screen_height - height);
+	counter = (int)(pixels_at_speed) / (screen_height - height);
+	pixel_direction = counter % 2;
+
+	if (!finished_update) {
+		if ((int)(pixels_at_speed) > (zoom_region->roi.y1 + height))
+			roi.y1 = zoom_region->roi.y1 + height;
+		else
+			roi.y1 = (int)(pixels_at_speed);
+
+		if (roi.y1 >= screen_height - height) {
+			roi.y1 = screen_height - height;
+			reset_timing = TRUE;
+		}
+	} else {
+		if (pixel_direction == 0)
+			roi.y1 = screen_height - height - pixel_position;
+		else
+			roi.y1 = pixel_position;
+	}
+
+	roi.y2 = roi.y1 + height;
+	magnifier->priv->cursor_x = (roi.x2 + roi.x1) / 2;
+	magnifier->priv->cursor_y = (roi.y2 + roi.y1) / 2;
+
+	/* Add one since in first loop we call zoom_region_process_updates */
+	if (counter > zoom_region->timing_iterations - 1) {
+		fprintf(stderr, "\n  Pan Rate                 = %d\n", zoom_region->timing_pan_rate);
+		zoom_region->exit_magnifier = TRUE;
+	}
+
+	zoom_regions = GNOME_Magnifier_Magnifier_getZoomRegions (
+		BONOBO_OBJREF (magnifier), &ev);
+
+	if (zoom_regions && (zoom_regions->_length > 0)) {
+		GNOME_Magnifier_ZoomRegion_setROI (
+			zoom_regions->_buffer[0], &roi, &ev);
+	}
+
+	if (!finished_update) {
+		zoom_region_process_updates(zoom_region);
+		if (roi.y1 >= screen_height - height) {
+			finished_update = TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 static void
 impl_zoom_region_set_roi (PortableServer_Servant servant,
 			  const GNOME_Magnifier_RectBounds *bounds,
@@ -1768,7 +2286,41 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
 	zoom_region->roi = *bounds;
-	zoom_region_align (zoom_region);
+
+	if (zoom_region->timing_pan_rate > 0) {
+		/* Set idle handler to do panning test */
+		g_idle_add_full (GDK_PRIORITY_REDRAW + 3, 
+			zoom_region_pan_test, zoom_region, NULL);
+	}
+
+	if (zoom_region->exit_magnifier) {
+		exit_report(zoom_region);
+	}
+
+	/*
+	 * Do not bother trying to update the screen if the last
+	 * screen update has not had time to complete.
+	 */
+	if (processing_updates) {
+		/* Remove any previous idle handler */
+		if (pending_idle_handler != 0) {
+			g_source_remove(pending_idle_handler);
+			pending_idle_handler = 0;
+		}
+
+		/* Set idle handler to process this pending update when possible */
+		pending_idle_handler = g_idle_add_full (GDK_PRIORITY_REDRAW + 2,
+			zoom_region_process_pending, zoom_region, NULL);
+
+		if (zoom_region->timing_output) {
+			fprintf(stderr,
+				"\n  [Last update not finished, pending - ROI=(%d, %d) (%d, %d)]\n\n",
+				zoom_region->roi.x1, zoom_region->roi.y1, zoom_region->roi.x2,
+				zoom_region->roi.y2);
+		}
+	} else {
+		zoom_region_align (zoom_region);
+	}
 }
 
 static CORBA_boolean
@@ -1924,20 +2476,22 @@ zoom_region_dispose (GObject *object)
 static void
 zoom_region_class_init (ZoomRegionClass *klass)
 {
-        GObjectClass * object_class = (GObjectClass *) klass;
-        POA_GNOME_Magnifier_ZoomRegion__epv *epv = &klass->epv;
+	GObjectClass * object_class = (GObjectClass *) klass;
+	POA_GNOME_Magnifier_ZoomRegion__epv *epv = &klass->epv;
 
 	object_class->dispose = zoom_region_dispose;
-        object_class->finalize = zoom_region_finalize;
+	object_class->finalize = zoom_region_finalize;
 	
-        epv->setMagFactor = impl_zoom_region_set_mag_factor;
-        epv->getMagFactor = impl_zoom_region_get_mag_factor;
-        epv->getProperties = impl_zoom_region_get_properties;
+	epv->setMagFactor = impl_zoom_region_set_mag_factor;
+	epv->getMagFactor = impl_zoom_region_get_mag_factor;
+	epv->getProperties = impl_zoom_region_get_properties;
 	epv->setROI = impl_zoom_region_set_roi;
-        epv->markDirty = impl_zoom_region_mark_dirty;
+	epv->markDirty = impl_zoom_region_mark_dirty;
 	epv->getROI = impl_zoom_region_get_roi;
 	epv->moveResize = impl_zoom_region_move_resize;
 	epv->dispose = impl_zoom_region_dispose;
+
+	reset_timing_stats();
 }
 
 static void
@@ -2116,6 +2670,60 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
 
+	def = bonobo_arg_new (BONOBO_ARG_INT);
+	BONOBO_ARG_SET_INT (def, 0);
+
+	bonobo_property_bag_add (zoom_region->properties,
+				 "timing-iterations",
+				 ZOOM_REGION_TIMING_TEST_PROP,
+				 BONOBO_ARG_INT,
+				 def,
+				 "timing iterations",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
+
+	def = bonobo_arg_new (BONOBO_ARG_BOOLEAN);
+	BONOBO_ARG_SET_BOOLEAN (def, FALSE);
+	
+	bonobo_property_bag_add (zoom_region->properties,
+				 "timing-output",
+				 ZOOM_REGION_TIMING_OUTPUT_PROP,
+				 BONOBO_ARG_BOOLEAN,
+				 def,
+				 "timing output",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+
+	bonobo_arg_release (def);
+
+	def = bonobo_arg_new (BONOBO_ARG_INT);
+	BONOBO_ARG_SET_INT (def, 0);
+
+	bonobo_property_bag_add (zoom_region->properties,
+				 "timing-pan-rate",
+				 ZOOM_REGION_TIMING_PAN_RATE_PROP,
+				 BONOBO_ARG_INT,
+				 def,
+				 "timing pan rate",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
+
+	def = bonobo_arg_new (BONOBO_ARG_BOOLEAN);
+	BONOBO_ARG_SET_BOOLEAN (def, FALSE);
+	
+	bonobo_property_bag_add (zoom_region->properties,
+				 "exit-magnifier",
+				 ZOOM_REGION_EXIT_MAGNIFIER,
+				 BONOBO_ARG_BOOLEAN,
+				 def,
+				 "timing output",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+
+	bonobo_arg_release (def);
+
 }
 
 static void
@@ -2159,6 +2767,7 @@ zoom_region_init (ZoomRegion *zoom_region)
 				     BONOBO_OBJECT (zoom_region->properties));
 	zoom_region->priv = g_malloc (sizeof (ZoomRegionPrivate));
 	zoom_region_private_init (zoom_region->priv);
+	zoom_region->timing_output = FALSE;
 	g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE,
 			    200,
 			    zoom_region_update_pointer_timeout,
